"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const yargs_1 = __importDefault(require("yargs"));
const lodash_1 = require("lodash");
const path_1 = __importDefault(require("path"));
const fs_1 = require("fs");
const util_1 = require("util");
const util_2 = require("./util");
const pReadFile = util_1.promisify(fs_1.readFile);
const getAuthConfig = async (credentialArgs) => {
    if (!credentialArgs) {
        return undefined;
    }
    const authConfig = new Map();
    await Promise.all(credentialArgs.map(async (credentials) => {
        const lastColon = credentials.lastIndexOf(":");
        if (lastColon === -1) {
            throw new Error(`Credentials ${lastColon} should be in format repository-url:file-path`);
        }
        const repoUrl = credentials.substr(0, lastColon);
        const credentialsFilePath = credentials.substr(lastColon + 1);
        const credentialsLine = await pReadFile(credentialsFilePath);
        const [username, password] = credentialsLine
            .toString()
            .trim()
            .split(":");
        if (!username || !password) {
            throw new Error("Credential file must have the format `username:password'");
        }
        authConfig.set(repoUrl, { username, password });
    }));
    util_2.debugLog("Read credentials %O", authConfig);
    return authConfig;
};
function isSubDirectory(referencePath, childPath) {
    return !path_1.default.relative(referencePath, childPath).startsWith("../");
}
exports.getConfigFromArgv = async () => {
    const argBuilder = yargs_1.default
        .option("credentials", {
        describe: "repo url followed by colon followed by path to an file specifying the colon separated container registry credentials",
        alias: "c",
        type: "array",
        string: true,
    })
        .option("images", {
        describe: "comma separated list of images to watch for updates",
        type: "string",
        alias: "i",
    })
        .option("shutdown-timeout", {
        describe: "how long to wait in seconds for docker container to shut down",
        alias: "s",
        type: "number",
    })
        .option("directory", {
        describe: "directory containing docker-compose.yml",
        alias: "d",
        type: "string",
    })
        .option("debounce", {
        describe: "number of seconds to debounce detected tags before restarting docker-compose",
        alias: "D",
        type: "number",
        default: 60,
    })
        .option("poll-interval", {
        describe: "how often to poll container repositories for updates in seconds",
        alias: "I",
        type: "number",
        default: 60,
    })
        .option("health-check-interval", {
        describe: "how often to check for health of docker containers in seconds",
        alias: "H",
        type: "number",
        default: 60,
    })
        .option("ill-health-tolerance", {
        describe: "how many seconds a container is allowed to be unhealthy before it is terminated",
        alias: "t",
        type: "number",
        default: 60,
    })
        .option("insecure", {
        describe: "Allow insecure HTTPS certificates for container registry",
        alias: "k",
        type: "boolean",
    })
        .option("signal-poll", {
        describe: "Send signal to existing friendship-blaster to force an update poll",
        alias: "S",
        type: "boolean",
    })
        .alias("h", "help")
        .help();
    const rawConfig = argBuilder.strict().argv;
    const { directory = process.cwd() } = rawConfig;
    const { credentials } = rawConfig;
    if (credentials) {
        credentials.forEach(oneCredential => {
            if (!isSubDirectory(directory, oneCredential)) {
                throw new Error(`${oneCredential} must be within the directory ${directory}`);
            }
        });
    }
    return Object.freeze({
        images: Object.freeze(new Set((rawConfig.images || "").split(/\s*,\s*/))),
        shutdownTimeout: rawConfig["shutdown-timeout"] || 10,
        directory,
        pollInterval: rawConfig["poll-interval"],
        debounce: rawConfig.debounce,
        allowInsecureHttps: !!rawConfig.insecure,
        healthCheckInterval: rawConfig["health-check-interval"],
        illHealthTolerance: rawConfig["ill-health-tolerance"],
        auth: await getAuthConfig(credentials),
        signalPoll: !!rawConfig["signal-poll"],
    });
};
exports.getVersion = lodash_1.once(() => {
    return require(path_1.default.join(__dirname, "..", "package.json")).version;
});
//# sourceMappingURL=config.js.map