"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const dockerode_1 = __importDefault(require("dockerode"));
const operators_1 = require("rxjs/operators");
const util_1 = require("./util");
const processes_1 = require("./processes");
function assertTaggedImageList(thing) {
    if (!Array.isArray(thing)) {
        throw new Error("Version file should contain a list");
    }
    thing.forEach(dockerImage => {
        if (typeof dockerImage.image !== "string") {
            throw new Error("Version entry did not contain string image property");
        }
        if (typeof dockerImage.tag !== "string") {
            throw new Error("Version entry did not contain string tag property");
        }
        if (typeof dockerImage.repoUrl !== "string") {
            throw new Error("Version entry did not contain string repoUrl property");
        }
    });
}
exports.assertTaggedImageList = assertTaggedImageList;
async function loginToContainerRepositories(authConfig) {
    await Promise.all(Array.from(authConfig.keys()).map(async (repoUrl) => {
        const repoAuth = authConfig.get(repoUrl);
        util_1.debugLog("Logging in to container registry", repoUrl);
        await processes_1.runCommand([
            "docker",
            "login",
            repoUrl,
            "-u",
            repoAuth.username,
            "-p",
            repoAuth.password,
        ]);
        util_1.debugLog("Logged in to container registry", repoUrl);
    }));
}
exports.loginToContainerRepositories = loginToContainerRepositories;
function assertDockerComposeConfig(thing) {
    var _a;
    if (typeof thing !== "object") {
        throw new Error("Invalid docker-compose.yml");
    }
    const services = (_a = thing) === null || _a === void 0 ? void 0 : _a.services;
    if (!services) {
        throw new Error("Invalid docker-compose.yml: did not contain services");
    }
    return Object.values(services).forEach(service => {
        if (typeof service !== "object") {
            throw new Error("Invalid service found in docker-compose.yml");
        }
        const imageStr = service.image;
        if (!imageStr) {
            throw new Error("Invalid service image found in docker-compose.yml");
        }
    });
}
exports.assertDockerComposeConfig = assertDockerComposeConfig;
exports.waitForDocker = (docker, stream) => new Promise(resolve => docker.modem.followProgress(stream, resolve));
const getAuthOptions = (repoUrl, auth) => {
    const repoAuth = auth === null || auth === void 0 ? void 0 : auth.get(repoUrl);
    if (!repoAuth) {
        return {};
    }
    return {
        authconfig: { username: repoAuth.username, password: repoAuth.password },
    };
};
async function pullChangedImages(auth, prevImages, newPollableImages) {
    const changedImages = newPollableImages.filter(({ repoUrl, image, tag }) => {
        const matchingPrev = prevImages.find(prev => prev.repoUrl == repoUrl && prev.image === image);
        return matchingPrev.tag !== tag;
    });
    const docker = new dockerode_1.default();
    util_1.debugLog("Pull changed images: %O", changedImages);
    await Promise.all(changedImages.map(async (image) => {
        const pullStream = await docker.pull(`${image.repoUrl}/${image.image}:${image.tag}`, getAuthOptions(image.repoUrl, auth));
        await exports.waitForDocker(docker, pullStream);
    }));
    util_1.debugLog("Pulled changed images");
}
exports.pullChangedImages = pullChangedImages;
exports.restartVeryUnhealthyContainers = (shutdownTimeout) => {
    const pendingRestarts = new Set();
    return operators_1.mergeMap(async (containerLabel) => {
        if (pendingRestarts.has(containerLabel)) {
            return;
        }
        pendingRestarts.add(containerLabel);
        try {
            console.warn("Restarting unhealthy container: %s", containerLabel);
            await processes_1.runCommand([
                "docker-compose",
                "restart",
                "-t",
                shutdownTimeout.toString(),
                containerLabel,
            ]);
            console.warn("Restarted unhealthy container: %s", containerLabel);
        }
        finally {
            pendingRestarts.delete(containerLabel);
        }
    });
};
//# sourceMappingURL=docker.js.map