"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const dockerode_1 = __importDefault(require("dockerode"));
const delay_1 = __importDefault(require("delay"));
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const util_1 = require("./util");
const processes_1 = require("./processes");
const MILISECS_IN_A_SECOND = 1000;
const CONTAINER_ID_CHECK_INTERVAL = 1000;
const MONITOR_HEALTH_RECOVERY_INTERVAL = 10000;
async function getContainerIdFromDockerComposeLabel(label) {
    while (true) {
        try {
            const containerId = (await processes_1.execCommand(["docker-compose", "ps", "-q", label])).trimRight();
            if (containerId.length) {
                return containerId;
            }
        }
        catch (e) {
            console.warn("Found error while checking for container ID: %O", e);
        }
        await delay_1.default(CONTAINER_ID_CHECK_INTERVAL);
    }
}
const checkContainerDockerHealth = (containerStatus) => async () => {
    const docker = new dockerode_1.default();
    const containerId = containerStatus.containerId ||
        (await getContainerIdFromDockerComposeLabel(containerStatus.label));
    try {
        const inspection = await docker.getContainer(containerId).inspect();
        const health = inspection.State.Health;
        if (!health || health.Status !== "unhealthy") {
            return {
                ...containerStatus,
                containerId,
                lastHealthyTime: new Date(),
            };
        }
        else {
            return { ...containerStatus, containerId };
        }
    }
    catch (e) {
        if (e.reason === "no such container") {
            console.warn("Found stale container while performing healthcheck: %s", containerId);
            return {
                label: containerStatus.label,
                containerId: undefined,
                lastHealthyTime: new Date(),
            };
        }
        else {
            throw e;
        }
    }
};
function getVeryUnhealthyDockerContainers(healthPollFrequency, illHealthTolerance, dockerComposeConfig) {
    const initialContainerStatuses = Object.keys(dockerComposeConfig.services).map((label) => ({
        label,
        containerId: undefined,
        lastHealthyTime: new Date(),
    }));
    return rxjs_1.merge(initialContainerStatuses).pipe(operators_1.mergeMap(initialContainerStatus => {
        return rxjs_1.interval(healthPollFrequency * MILISECS_IN_A_SECOND).pipe(util_1.switchScan((containerStatus) => {
            return util_1.promiseFactoryToObservable(checkContainerDockerHealth(containerStatus)).pipe(util_1.logErrorAndRetry("Error monitoring docker container health, retrying: %O", MONITOR_HEALTH_RECOVERY_INTERVAL));
        }, initialContainerStatus), operators_1.map((containerStatus) => {
            const now = Date.now();
            if ((now - containerStatus.lastHealthyTime.getTime()) /
                MILISECS_IN_A_SECOND >
                illHealthTolerance) {
                return containerStatus.label;
            }
            else {
                return undefined;
            }
        }), operators_1.filter(util_1.isDefined));
    }));
}
exports.getVeryUnhealthyDockerContainers = getVeryUnhealthyDockerContainers;
//# sourceMappingURL=healthcheck.js.map