"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const yaml_1 = __importDefault(require("yaml"));
const path_1 = __importDefault(require("path"));
const util_1 = __importDefault(require("util"));
const md5_1 = __importDefault(require("md5"));
const operators_1 = require("rxjs/operators");
const rxjs_1 = require("rxjs");
const fs_1 = require("fs");
const lodash_1 = require("lodash");
const config_1 = require("./config");
const processes_1 = require("./processes");
const docker_1 = require("./docker");
const healthcheck_1 = require("./healthcheck");
const pollImagesForUpdate_1 = __importDefault(require("./pollImagesForUpdate"));
const util_2 = require("./util");
const pExists = util_1.default.promisify(fs_1.exists);
const pReadFile = util_1.default.promisify(fs_1.readFile);
const pWriteFile = util_1.default.promisify(fs_1.writeFile);
const DOCKER_COMPOSE_YML = "docker-compose.yml";
exports.FBLASTER_COMPOSE_FILE = "fblaster-docker-compose.yml";
exports.FBLASTER_VERSION_FILE = "fblaster-versions.yml";
const DOCKER_PULL_RETRY_INTERVAL = 3000;
const DOCKER_COMPOSE_RESTART_RETRY_INTERVAL = 3000;
async function runSignalPoll(directoryPath) {
    console.log("Signalling existing friendship-blaster to poll for updates");
    try {
        const containerName = `fblaster-${md5_1.default(directoryPath)}`;
        await processes_1.spawnProcess(["docker", "kill", "-s", "SIGUSR2", containerName], {
            showStderr: true,
        }).wait();
        console.log("Signalled existing friendship-blaster to poll for updates");
    }
    catch (e) {
        console.warn("Failed to signal existing friendship-blaster to poll for updates");
        process.exit(1);
    }
}
async function getPollableImages(dockerComposeConfig, pollSpec) {
    const pollableImages = Object.values(dockerComposeConfig.services)
        .map(service => {
        if (typeof service !== "object") {
            throw new Error(`Invalid service found in ${DOCKER_COMPOSE_YML}`);
        }
        const imageStr = service.image;
        if (!imageStr) {
            return undefined;
        }
        const [repoUrl, imageAndTag] = imageStr.split("/");
        if (!repoUrl || !imageAndTag) {
            return undefined;
        }
        const [image, tag] = imageAndTag.split(":");
        if (!image || !tag) {
            return undefined;
        }
        const repoAndImage = `${repoUrl}/${image}`;
        return pollSpec.has(repoAndImage) || pollSpec.has(image)
            ? { repoUrl, image, tag }
            : undefined;
    })
        .filter(util_2.isDefined);
    return Object.freeze(pollableImages);
}
async function writeFblasterDockerCompose(filePath, dockerComposeConfig) {
    const yamlString = yaml_1.default.stringify(dockerComposeConfig);
    await pWriteFile(filePath, yamlString);
}
function subscribeToImageUpdates(pollableImages, config) {
    return pollImagesForUpdate_1.default(pollableImages, config.pollInterval, config.allowInsecureHttps, config.auth);
}
async function getDockerComposeConfig(directoryPath) {
    const composeFilePath = path_1.default.join(directoryPath, DOCKER_COMPOSE_YML);
    if (!(await pExists(composeFilePath))) {
        throw new Error(`${composeFilePath} must exit`);
    }
    const dockerComposeConfig = await yaml_1.default.parse((await pReadFile(composeFilePath)).toString());
    docker_1.assertDockerComposeConfig(dockerComposeConfig);
    return Object.freeze(dockerComposeConfig);
}
async function shutdownDockerCompose(proc, fblasterComposeConfigPath, shutdownTimeout) {
    try {
        await proc.shutdown();
    }
    catch (e) {
        console.warn("Error stopping docker-compose process: %O", e);
    }
    try {
        await processes_1.runCommand([
            "docker-compose",
            "-f",
            fblasterComposeConfigPath,
            "stop",
            "-t",
            shutdownTimeout.toString(),
        ]);
    }
    catch (e) {
        console.warn("Error running `docker-compose stop`: %O", e);
    }
}
function getDockerComposeSpawner(config) {
    const fblasterComposeConfigPath = path_1.default.join(config.directory, exports.FBLASTER_COMPOSE_FILE);
    let proc;
    let healthCheck;
    let procMonitor;
    const respawn = async (dockerComposeConfig) => {
        await writeFblasterDockerCompose(fblasterComposeConfigPath, dockerComposeConfig);
        if (healthCheck) {
            healthCheck.unsubscribe();
        }
        if (procMonitor) {
            procMonitor.unsubscribe();
        }
        if (proc) {
            const procCopy = proc;
            proc = undefined;
            await shutdownDockerCompose(procCopy, fblasterComposeConfigPath, config.shutdownTimeout);
        }
        proc = processes_1.spawnProcess([
            "docker-compose",
            "-f",
            fblasterComposeConfigPath,
            "up",
            "-t",
            config.shutdownTimeout.toString(),
        ]);
        healthCheck = healthcheck_1.getVeryUnhealthyDockerContainers(config.healthCheckInterval, config.illHealthTolerance, dockerComposeConfig)
            .pipe(docker_1.restartVeryUnhealthyContainers(config.shutdownTimeout))
            .subscribe();
        const restartExitedProcess = () => {
            proc = undefined;
            respawn(dockerComposeConfig);
        };
        procMonitor = rxjs_1.from(proc.wait()).subscribe(restartExitedProcess, restartExitedProcess);
    };
    return Object.freeze({
        respawn,
        async shutdown() {
            if (healthCheck) {
                healthCheck.unsubscribe();
            }
            if (procMonitor) {
                procMonitor.unsubscribe();
            }
            if (proc) {
                await shutdownDockerCompose(proc, fblasterComposeConfigPath, config.shutdownTimeout);
            }
        },
    });
}
function mergePollableImagesWithDockerComposeConfig(pollableImages, dockerComposeConfig) {
    const newConfig = { ...dockerComposeConfig };
    newConfig.services = lodash_1.mapValues(newConfig.services, service => {
        const [repoUrl, imageAndTag] = service.image.split("/");
        if (!repoUrl || !imageAndTag) {
            return service;
        }
        const [image, tag] = imageAndTag.split(":");
        if (!image || !tag) {
            return service;
        }
        const matchingImage = pollableImages.find(pollableImage => pollableImage.image === image && pollableImage.repoUrl === repoUrl);
        if (matchingImage) {
            return {
                ...service,
                image: `${repoUrl}/${image}:${matchingImage.tag}`,
            };
        }
        else {
            return service;
        }
    });
    return newConfig;
}
async function respawnDockerComposeWithNewConfig(pollableImages, dockerComposeConfig, respawnDockerCompose) {
    util_2.debugLog("Restart docker-compose process with new config: %O - %O", pollableImages, dockerComposeConfig);
    const newConfig = mergePollableImagesWithDockerComposeConfig(pollableImages, dockerComposeConfig);
    await respawnDockerCompose(newConfig);
}
const tryToPullImages = (auth, [prevImages, newPollableImages]) => {
    return util_2.promiseFactoryToObservable(async () => {
        await docker_1.pullChangedImages(auth, prevImages, newPollableImages);
        return newPollableImages;
    }).pipe(util_2.logErrorAndRetry("Error pulling changed images, retrying: %O", DOCKER_PULL_RETRY_INTERVAL));
};
const tryToRestartDockerCompose = (dockerComposeConfig, respawnDockerCompose, newPollableImages) => {
    return util_2.promiseFactoryToObservable(async () => {
        await respawnDockerComposeWithNewConfig(newPollableImages, dockerComposeConfig, respawnDockerCompose);
        return {
            images: newPollableImages,
            config: dockerComposeConfig,
        };
    }).pipe(util_2.logErrorAndRetry("Error restarting docker, retrying: %O", DOCKER_COMPOSE_RESTART_RETRY_INTERVAL));
};
async function writeVersionFile(directoryPath, pollableImages) {
    const versionFilePath = path_1.default.join(directoryPath, exports.FBLASTER_VERSION_FILE);
    const yamlString = yaml_1.default.stringify(pollableImages);
    await pWriteFile(versionFilePath, yamlString);
}
async function getLatestPollableImages(directoryPath, pollableImages) {
    const versionFilePath = path_1.default.join(directoryPath, exports.FBLASTER_VERSION_FILE);
    if (!(await pExists(versionFilePath))) {
        return pollableImages;
    }
    const latestImages = await yaml_1.default.parse((await pReadFile(versionFilePath)).toString());
    docker_1.assertTaggedImageList(latestImages);
    return pollableImages.map(pollableImage => {
        const matching = latestImages.find(latestImage => latestImage.repoUrl === pollableImage.repoUrl &&
            latestImage.image === pollableImage.image);
        return matching ? { ...pollableImage, tag: matching.tag } : pollableImage;
    });
}
async function runFriendshipBlaster(config) {
    if (config.auth) {
        await docker_1.loginToContainerRepositories(config.auth);
    }
    const pureDockerComposeConfig = await getDockerComposeConfig(config.directory);
    const initialPollableImages = await getPollableImages(pureDockerComposeConfig, config.images);
    const pollableImages = await getLatestPollableImages(config.directory, initialPollableImages);
    const dockerComposeConfig = pollableImages === initialPollableImages
        ? pureDockerComposeConfig
        : mergePollableImagesWithDockerComposeConfig(pollableImages, pureDockerComposeConfig);
    const dockerComposeSpawner = getDockerComposeSpawner(config);
    dockerComposeSpawner.respawn(dockerComposeConfig);
    console.info("Run initial images: %O", pollableImages);
    const pollObservable = subscribeToImageUpdates(pollableImages, config);
    const pollSubscription = pollObservable
        .pipe(operators_1.startWith(pollableImages), operators_1.debounceTime(config.debounce * 1000), operators_1.pairwise(), operators_1.switchMap(tryToPullImages.bind(null, config.auth)), operators_1.switchMap(tryToRestartDockerCompose.bind(null, dockerComposeConfig, dockerComposeSpawner.respawn)), operators_1.switchMap(async (imagesAndConfig) => {
        await writeVersionFile(config.directory, imagesAndConfig.images);
        return imagesAndConfig.images;
    }))
        .subscribe((pollableImages) => {
        console.info("Updated to latest images: %O", pollableImages);
    }, (error) => {
        console.error("Error running subscription poller: %O", error ? error.message : error);
    });
    return () => {
        dockerComposeSpawner.shutdown();
        pollSubscription.unsubscribe();
    };
}
exports.default = runFriendshipBlaster;
async function main() {
    try {
        const config = await config_1.getConfigFromArgv();
        if (config.signalPoll) {
            await runSignalPoll(config.directory);
        }
        else {
            if (!config.images.size) {
                console.warn("Must provide one of --images/-i or --signal-poll/-S");
                process.exit(1);
            }
            const onExit = lodash_1.once(await runFriendshipBlaster(config));
            process.on("exit", onExit);
            process.on("SIGINT", onExit);
            process.on("SIGTERM", onExit);
            process.on("uncaughtException", onExit);
        }
    }
    catch (e) {
        console.warn(e.message);
        process.exit(1);
    }
}
if (require.main === module) {
    main();
}
//# sourceMappingURL=index.js.map