"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const delay_1 = __importDefault(require("delay"));
const fs_1 = require("fs");
const util_1 = require("util");
const lodash_1 = require("lodash");
const testUtils_1 = require("./testUtils");
const processes_1 = require("./processes");
const _1 = require(".");
const pUnlink = util_1.promisify(fs_1.unlink);
describe("friendship-blaster", () => {
    jest.setTimeout(70000);
    let fblasterProcess = undefined;
    const shutdownFblaster = async () => {
        if (fblasterProcess) {
            const procCopy = fblasterProcess;
            fblasterProcess = undefined;
            await procCopy.shutdown();
        }
    };
    afterEach(() => Promise.all([testUtils_1.stopTestContainerRegistry(), shutdownFblaster()]));
    describe("with simple configuration", () => {
        const ECHO_SERVER = "echo-server";
        const TYPE1 = "type1";
        const TYPE2 = "type2";
        const INITIAL_VERSION_TYPE1 = "10.0.0";
        const INITIAL_VERSION_TYPE2 = "8.0.0";
        const UPDATED_VERSION_TYPE1 = "10.0.1";
        const UPDATED_VERSION_TYPE2 = "8.0.1";
        const spawnFriendshipBlasterWithSimpleConfig = (withAuth, pollInterval = 5) => {
            fblasterProcess = testUtils_1.spawnTestFriendshipBlaster("echo-server-type1,echo-server-type2", { withAuth, pollInterval });
        };
        const waitForInitialState = async () => {
            const [lines1, lines2] = await Promise.all([
                testUtils_1.pollFileForLines("mnt/type1", 1),
                testUtils_1.pollFileForLines("mnt/type2", 1),
            ]);
            expect(lines1).toEqual([INITIAL_VERSION_TYPE1]);
            expect(lines2).toEqual([INITIAL_VERSION_TYPE2]);
        };
        const prepareTestConfigurationDirectory = async () => {
            await testUtils_1.chdirToTestConfig("simple");
            await Promise.all([
                pUnlink(_1.FBLASTER_VERSION_FILE).catch(() => { }),
                testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE1, INITIAL_VERSION_TYPE1),
                testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE2, INITIAL_VERSION_TYPE2),
                testUtils_1.emptyDirectory("mnt"),
            ]);
        };
        afterEach(async () => {
            await processes_1.runCommand(["docker-compose", "down"]);
            await Promise.all([
                testUtils_1.removeTestImage(ECHO_SERVER, TYPE1, INITIAL_VERSION_TYPE1),
                testUtils_1.removeTestImage(ECHO_SERVER, TYPE2, INITIAL_VERSION_TYPE2),
                testUtils_1.removeTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1),
                testUtils_1.removeTestImage(ECHO_SERVER, TYPE2, UPDATED_VERSION_TYPE2),
            ]);
        });
        describe("and no authentication", () => {
            beforeEach(async () => {
                await testUtils_1.startTestContainerRegistry(false);
                await prepareTestConfigurationDirectory();
                spawnFriendshipBlasterWithSimpleConfig(false);
            });
            it("is able to run simple docker-compose.yml configuration", async () => {
                await waitForInitialState();
            });
        });
        describe("and authentication", () => {
            beforeEach(async () => {
                await testUtils_1.startTestContainerRegistry(true);
                await testUtils_1.loginToTestContainerRegistry();
                await prepareTestConfigurationDirectory();
                spawnFriendshipBlasterWithSimpleConfig(true);
            });
            it("is able to run simple docker-compose.yml configuration", async () => {
                await waitForInitialState();
            });
            it("is able to poll for updates and restart configuration when a single update is found", async () => {
                await waitForInitialState();
                await testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1);
                await testUtils_1.removeTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1);
                const [lines1, lines2] = await Promise.all([
                    testUtils_1.pollFileForLines("mnt/type1", 2),
                    testUtils_1.pollFileForLines("mnt/type2", 2),
                ]);
                expect(lines1).toEqual([INITIAL_VERSION_TYPE1, UPDATED_VERSION_TYPE1]);
                expect(lines2).toEqual([INITIAL_VERSION_TYPE2, INITIAL_VERSION_TYPE2]);
            });
            it("skips incompatible versions", async () => {
                await waitForInitialState();
                const stupidlyHighVersion = "400.0.0";
                await testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE1, stupidlyHighVersion);
                await delay_1.default(20000);
                await testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1);
                await testUtils_1.removeTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1);
                const [lines1, lines2] = await Promise.all([
                    testUtils_1.pollFileForLines("mnt/type1", 2),
                    testUtils_1.pollFileForLines("mnt/type2", 2),
                ]);
                expect(lines1).toEqual([INITIAL_VERSION_TYPE1, UPDATED_VERSION_TYPE1]);
                expect(lines2).toEqual([INITIAL_VERSION_TYPE2, INITIAL_VERSION_TYPE2]);
                await shutdownFblaster();
                await testUtils_1.removeTestImage(ECHO_SERVER, TYPE1, stupidlyHighVersion);
            });
            it("debounces multiple changes to avoid unnecessary restarts", async () => {
                await waitForInitialState();
                await Promise.all([
                    testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1),
                    testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE2, UPDATED_VERSION_TYPE2),
                ]);
                await Promise.all([
                    testUtils_1.removeTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1),
                    testUtils_1.removeTestImage(ECHO_SERVER, TYPE2, UPDATED_VERSION_TYPE2),
                ]);
                const [lines1, lines2] = await Promise.all([
                    testUtils_1.pollFileForLines("mnt/type1", 2),
                    testUtils_1.pollFileForLines("mnt/type2", 2),
                ]);
                expect(lines1).toEqual([INITIAL_VERSION_TYPE1, UPDATED_VERSION_TYPE1]);
                expect(lines2).toEqual([INITIAL_VERSION_TYPE2, UPDATED_VERSION_TYPE2]);
            });
            it("responds to multiple changes outside of debounce period", async () => {
                await waitForInitialState();
                await testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1);
                await testUtils_1.removeTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1);
                const [firstLines1, firstLines2] = await Promise.all([
                    testUtils_1.pollFileForLines("mnt/type1", 2),
                    testUtils_1.pollFileForLines("mnt/type2", 2),
                ]);
                expect(firstLines1).toEqual([
                    INITIAL_VERSION_TYPE1,
                    UPDATED_VERSION_TYPE1,
                ]);
                expect(firstLines2).toEqual([
                    INITIAL_VERSION_TYPE2,
                    INITIAL_VERSION_TYPE2,
                ]);
                await testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE2, UPDATED_VERSION_TYPE2);
                await testUtils_1.removeTestImage(ECHO_SERVER, TYPE2, UPDATED_VERSION_TYPE2);
                const [lines1, lines2] = await Promise.all([
                    testUtils_1.pollFileForLines("mnt/type1", 3),
                    testUtils_1.pollFileForLines("mnt/type2", 3),
                ]);
                expect(lines1).toEqual([
                    INITIAL_VERSION_TYPE1,
                    UPDATED_VERSION_TYPE1,
                    UPDATED_VERSION_TYPE1,
                ]);
                expect(lines2).toEqual([
                    INITIAL_VERSION_TYPE2,
                    INITIAL_VERSION_TYPE2,
                    UPDATED_VERSION_TYPE2,
                ]);
            });
            it("picks up latest image versions when restarted", async () => {
                await waitForInitialState();
                await Promise.all([
                    testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1),
                    testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE2, UPDATED_VERSION_TYPE2),
                ]);
                await Promise.all([
                    testUtils_1.removeTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1),
                    testUtils_1.removeTestImage(ECHO_SERVER, TYPE2, UPDATED_VERSION_TYPE2),
                ]);
                const [preRestartLines1, preRestartLines2] = await Promise.all([
                    testUtils_1.pollFileForLines("mnt/type1", 2),
                    testUtils_1.pollFileForLines("mnt/type2", 2),
                ]);
                expect(preRestartLines1).toEqual([
                    INITIAL_VERSION_TYPE1,
                    UPDATED_VERSION_TYPE1,
                ]);
                expect(preRestartLines2).toEqual([
                    INITIAL_VERSION_TYPE2,
                    UPDATED_VERSION_TYPE2,
                ]);
                await shutdownFblaster();
                spawnFriendshipBlasterWithSimpleConfig(true);
                const [lines1, lines2] = await Promise.all([
                    testUtils_1.pollFileForLines("mnt/type1", 3),
                    testUtils_1.pollFileForLines("mnt/type2", 3),
                ]);
                expect(lines1).toEqual([
                    INITIAL_VERSION_TYPE1,
                    UPDATED_VERSION_TYPE1,
                    UPDATED_VERSION_TYPE1,
                ]);
                expect(lines2).toEqual([
                    INITIAL_VERSION_TYPE2,
                    UPDATED_VERSION_TYPE2,
                    UPDATED_VERSION_TYPE2,
                ]);
            });
            it("restarts docker-compose when it exits unexpectedly", async () => {
                await waitForInitialState();
                const pid = await testUtils_1.getMatchingProc(args => args.length > 1 &&
                    /f(?:riendship-)?blaster\/(?:.*\/)?dist\/index.js$/.test(args[1]));
                expect(pid).not.toEqual(0);
                const dockerComposePid = await testUtils_1.getMatchingDescendentProc(pid, args => args.length > 1 && /docker-compose$/.test(args[1]));
                expect(dockerComposePid).not.toEqual(0);
                try {
                    process.kill(dockerComposePid);
                }
                catch (e) {
                    await processes_1.runCommand(["sudo", "kill", dockerComposePid.toString()]);
                }
                const [lines1, lines2] = await Promise.all([
                    testUtils_1.pollFileForLines("mnt/type1", 2),
                    testUtils_1.pollFileForLines("mnt/type2", 2),
                ]);
                expect(lines1).toEqual([INITIAL_VERSION_TYPE1, INITIAL_VERSION_TYPE1]);
                expect(lines2).toEqual([INITIAL_VERSION_TYPE2, INITIAL_VERSION_TYPE2]);
            });
        });
        describe("responds to SIGUSR2 signal", () => {
            beforeEach(async () => {
                await testUtils_1.startTestContainerRegistry(false);
                await prepareTestConfigurationDirectory();
                spawnFriendshipBlasterWithSimpleConfig(false, 999999);
            });
            it("by polling for updates", async () => {
                if (process.env.LOCAL_COMPOSE) {
                    console.warn("SIGUSR2 tests do not work when LOCAL_COMPOSE is set");
                    return;
                }
                await waitForInitialState();
                await testUtils_1.buildAndPushTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1);
                await testUtils_1.removeTestImage(ECHO_SERVER, TYPE1, UPDATED_VERSION_TYPE1);
                await delay_1.default(2000);
                await testUtils_1.spawnFriendshipBlasterSignaller();
                const [lines1, lines2] = await Promise.all([
                    testUtils_1.pollFileForLines("mnt/type1", 2),
                    testUtils_1.pollFileForLines("mnt/type2", 2),
                ]);
                expect(lines1).toEqual([INITIAL_VERSION_TYPE1, UPDATED_VERSION_TYPE1]);
                expect(lines2).toEqual([INITIAL_VERSION_TYPE2, INITIAL_VERSION_TYPE2]);
            }, 150000);
        });
    });
    describe("with healthcheck configuration", () => {
        const HEALTH_SERVER = "unhealthy-server";
        const VARIANT1 = "2";
        const VARIANT2 = "3";
        const INITIAL_VERSION_VARIANT1 = "1.0.0";
        const INITIAL_VERSION_VARIANT2 = "1.0.0";
        const UPDATED_VERSION_VARIANT1 = "1.0.1";
        const waitForInitialState = async () => {
            const [lines1, lines2] = await Promise.all([
                testUtils_1.pollFileForLines(`mnt/${VARIANT1}`, 1),
                testUtils_1.pollFileForLines(`mnt/${VARIANT2}`, 1),
            ]);
            expect(lines1).toEqual([INITIAL_VERSION_VARIANT1]);
            expect(lines2).toEqual([INITIAL_VERSION_VARIANT2]);
        };
        const prepareTestConfigurationDirectory = async () => {
            await testUtils_1.chdirToTestConfig("health");
            await Promise.all([
                pUnlink(_1.FBLASTER_VERSION_FILE).catch(() => { }),
                testUtils_1.buildAndPushTestImage(HEALTH_SERVER, VARIANT1, INITIAL_VERSION_VARIANT1),
                testUtils_1.buildAndPushTestImage(HEALTH_SERVER, VARIANT2, INITIAL_VERSION_VARIANT2),
                testUtils_1.emptyDirectory("mnt"),
            ]);
        };
        const spawnFriendshipBlasterWithHealthConfig = () => {
            fblasterProcess = testUtils_1.spawnTestFriendshipBlaster(`unhealthy-server-${VARIANT1},unhealthy-server-${VARIANT2}`, { withAuth: false });
        };
        beforeEach(async () => {
            await testUtils_1.startTestContainerRegistry(false);
            await prepareTestConfigurationDirectory();
            spawnFriendshipBlasterWithHealthConfig();
        });
        afterEach(async () => {
            await shutdownFblaster();
            await Promise.all([
                processes_1.runCommand(["docker-compose", "down"]),
                testUtils_1.removeTestImage(HEALTH_SERVER, VARIANT1, INITIAL_VERSION_VARIANT1),
                testUtils_1.removeTestImage(HEALTH_SERVER, VARIANT2, INITIAL_VERSION_VARIANT2),
                testUtils_1.removeTestImage(HEALTH_SERVER, VARIANT1, UPDATED_VERSION_VARIANT1),
            ]);
        });
        const waitForServersToGetHealthy = async () => {
            await waitForInitialState();
            const [lines1, lines2] = await Promise.all([
                testUtils_1.pollFileForLines(`mnt/${VARIANT1}`, 2, true),
                testUtils_1.pollFileForLines(`mnt/${VARIANT2}`, 3, true),
            ]);
            expect(lines1).toEqual(lodash_1.fill(new Array(2), INITIAL_VERSION_VARIANT1));
            expect(lines2).toEqual(lodash_1.fill(new Array(3), INITIAL_VERSION_VARIANT2));
        };
        it("restarts servers until they are healthy", async () => {
            await waitForServersToGetHealthy();
        });
        it("responds to image update after restarting unhealthy servers", async () => {
            await waitForServersToGetHealthy();
            await testUtils_1.buildAndPushTestImage(HEALTH_SERVER, VARIANT1, UPDATED_VERSION_VARIANT1);
            const [lines1] = await Promise.all([
                testUtils_1.pollFileForLines(`mnt/${VARIANT1}`, 3, true),
                testUtils_1.pollFileForLines(`mnt/${VARIANT2}`, 3, true),
            ]);
            expect(lines1).toEqual([
                INITIAL_VERSION_VARIANT1,
                INITIAL_VERSION_VARIANT1,
                UPDATED_VERSION_VARIANT1,
            ]);
        });
        it("restarts updated unhealthy servers", async () => {
            await waitForServersToGetHealthy();
            await testUtils_1.buildAndPushTestImage(HEALTH_SERVER, VARIANT1, UPDATED_VERSION_VARIANT1);
            await Promise.all([
                testUtils_1.pollFileForLines(`mnt/${VARIANT1}`, 3, true),
                testUtils_1.pollFileForLines(`mnt/${VARIANT2}`, 3, true),
            ]);
            await pUnlink(`mnt/${VARIANT1}`);
            const [lines1, lines2] = await Promise.all([
                testUtils_1.pollFileForLines(`mnt/${VARIANT1}`, 2, true),
                testUtils_1.pollFileForLines(`mnt/${VARIANT2}`, 4, true),
            ]);
            expect(lines1).toEqual(lodash_1.fill(new Array(2), UPDATED_VERSION_VARIANT1));
            expect(lines2).toEqual(lodash_1.fill(new Array(4), INITIAL_VERSION_VARIANT2));
        }, 130000);
    });
});
//# sourceMappingURL=index.spec.js.map