"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const rimraf_1 = __importDefault(require("rimraf"));
const util_1 = __importDefault(require("util"));
const delay_1 = __importDefault(require("delay"));
const fs_1 = require("fs");
const dockerode_1 = __importDefault(require("dockerode"));
const processes_1 = require("./processes");
const _1 = require(".");
const pRimraf = util_1.default.promisify(rimraf_1.default);
const pMkdir = util_1.default.promisify(fs_1.mkdir);
const pExists = util_1.default.promisify(fs_1.exists);
const pReadFile = util_1.default.promisify(fs_1.readFile);
const pUnlink = util_1.default.promisify(fs_1.unlink);
const TEST_CONTAINER_REPO_PORT = "5000";
const TEST_CONTAINER_REPO = `localhost:${TEST_CONTAINER_REPO_PORT}`;
const TEST_CONTAINER_REPO_USER = "testuser";
const TEST_CONTAINER_REPO_PASSWORD = "testpassword";
const REGISTRY_CONTAINER_NAME = "friendship-blaster-test-registry";
const FRIENDSHIP_BLASTER_BIN = path_1.default.join(__dirname, "..", "bin", "friendship-blaster");
const REGISTRY_CONTAINER_WAIT_DURATION = 10000;
const TEST_DOCKER_COMPOSE_CONFIG = path_1.default.join(__dirname, "..", "test", "docker-compose.yml");
const TEST_DOCKER_COMPOSE_WITH_AUTH_CONFIG = path_1.default.join(__dirname, "..", "test", "docker-compose-with-auth.yml");
const createDocker = () => new dockerode_1.default({ socketPath: "/var/run/docker.sock" });
let testDockerComposeProc = undefined;
exports.startTestContainerRegistry = async (withAuth) => {
    const authArgs = withAuth ? ["-f", TEST_DOCKER_COMPOSE_WITH_AUTH_CONFIG] : [];
    await processes_1.runCommand([
        "docker-compose",
        "-f",
        TEST_DOCKER_COMPOSE_CONFIG,
        ...authArgs,
        "down",
    ]);
    testDockerComposeProc = processes_1.spawnProcess([
        "docker-compose",
        "-f",
        TEST_DOCKER_COMPOSE_CONFIG,
        ...authArgs,
        "up",
        "-t",
        "2",
    ]);
    const docker = createDocker();
    const started = Date.now();
    const compareName = `/${REGISTRY_CONTAINER_NAME}`;
    do {
        const containers = await docker.listContainers();
        const [registryContainer] = containers.filter(container => container.Names.some(name => name === compareName));
        if (registryContainer && /^Up/.test(registryContainer.Status)) {
            return;
        }
        await delay_1.default(200);
    } while (Date.now() < started + REGISTRY_CONTAINER_WAIT_DURATION);
    throw new Error("Registry container did not start in time");
};
exports.loginToTestContainerRegistry = async () => processes_1.runCommand([
    "docker",
    "login",
    TEST_CONTAINER_REPO,
    "-u",
    TEST_CONTAINER_REPO_USER,
    "-p",
    TEST_CONTAINER_REPO_PASSWORD,
]);
exports.stopTestContainerRegistry = async () => {
    if (testDockerComposeProc) {
        const proc = testDockerComposeProc;
        testDockerComposeProc = undefined;
        await proc.shutdown();
    }
};
const buildImageName = (baseName, variant, tag) => `${TEST_CONTAINER_REPO}/${baseName}-${variant}:${tag}`;
exports.buildAndPushTestImage = async (baseName, variant, tag) => {
    const imageName = buildImageName(baseName, variant, tag);
    const directory = path_1.default.join(__dirname, "..", "test", "containers", baseName);
    await processes_1.runCommand([
        "docker",
        "build",
        "-t",
        imageName,
        "--build-arg",
        `IMAGE_TAG=${tag}`,
        "--build-arg",
        `VARIANT=${variant}`,
        directory,
    ]);
    await processes_1.runCommand(["docker", "push", imageName]);
};
exports.removeTestImage = async (baseName, variant, tag) => {
    const imageName = buildImageName(baseName, variant, tag);
    const docker = createDocker();
    try {
        await docker.getImage(imageName).remove();
    }
    catch (e) {
    }
};
exports.chdirToTestConfig = async (configName) => {
    process.chdir(path_1.default.join(__dirname, "..", "test", "configs", configName));
    try {
        await pUnlink(_1.FBLASTER_COMPOSE_FILE);
    }
    catch (e) { }
};
exports.emptyDirectory = async (dirPath) => {
    await pRimraf(dirPath);
    await pMkdir(dirPath);
};
exports.pollFileForLines = async (filePath, lineCount, exact = false) => {
    while (!(await pExists(filePath))) {
        await delay_1.default(100);
    }
    for (;;) {
        const lines = (await pReadFile(filePath))
            .toString()
            .trim()
            .split("\n");
        if (exact) {
            if (lines.length === lineCount) {
                return lines;
            }
            else if (lines.length > lineCount) {
                throw new Error(`Too many lines in ${filePath}, was expecting ${lineCount} and got ${lines.length}`);
            }
        }
        else {
            if (lines.length >= lineCount) {
                return lines;
            }
        }
        await delay_1.default(100);
    }
};
exports.spawnTestFriendshipBlaster = (imageArg, { withAuth, pollInterval = 5 }) => processes_1.spawnProcess([
    FRIENDSHIP_BLASTER_BIN,
    "--host-net",
    "--images",
    imageArg,
    "--shutdown-timeout",
    "2",
    "--poll-interval",
    pollInterval.toString(),
    "--debounce",
    "5",
    "--health-check-interval",
    "3",
    "--ill-health-tolerance",
    "6",
    ...(withAuth ? ["--credentials", "localhost:5000:credentials.txt"] : []),
    "--insecure",
], {
    showStderr: true,
});
exports.spawnFriendshipBlasterSignaller = () => processes_1.spawnProcess([FRIENDSHIP_BLASTER_BIN, "--signal-poll"], {
    showStderr: true,
}).wait();
exports.getMatchingProc = async (argMatcher) => {
    const psOutput = await processes_1.execCommand(["ps", "-eo", "pid,args"]);
    const matchingProcs = psOutput
        .trim()
        .split("\n")
        .slice(1)
        .map(line => {
        const [pidStr, ...args] = line.trim().split(/\s+/);
        return { pid: parseInt(pidStr), args };
    })
        .filter(({ args }) => argMatcher(args));
    return matchingProcs.length === 1 ? matchingProcs[0].pid : 0;
};
exports.getMatchingDescendentProc = async (parentPid, argMatcher) => {
    const psOutput = await processes_1.execCommand(["ps", "-eo", "pid,ppid,args"]);
    const pidMap = new Map();
    const matchingProcs = psOutput
        .trim()
        .split("\n")
        .slice(1)
        .map(line => {
        const [pidStr, ppidStr, ...args] = line.trim().split(/\s+/);
        const pid = parseInt(pidStr);
        const ppid = parseInt(ppidStr);
        pidMap.set(pid, ppid);
        return { pid, args };
    })
        .filter(pidInfo => {
        if (!argMatcher(pidInfo.args)) {
            return false;
        }
        for (let nextPid = pidInfo.pid; nextPid !== 0; nextPid = pidMap.get(nextPid) || 0) {
            if (nextPid === parentPid) {
                return true;
            }
        }
        return false;
    });
    return matchingProcs.length === 1 ? matchingProcs[0].pid : 0;
};
//# sourceMappingURL=testUtils.js.map